/*! \file   direct-motor.c
    \brief  direct motor access
    \author Markus L. Noga <noga@inrialpes.fr>
*/
    
/*
 *  The contents of this file are subject to the Mozilla Public License
 *  Version 1.0 (the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *  http://www.mozilla.org/MPL/
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *
 *  The Original Code is legOS code, released March 19, 1999.
 *
 *  The Initial Developer of the Original Code is Markus L. Noga.
 *  Portions created by Markus L. Noga are Copyright (C) 1999
 *  Markus L. Noga. All Rights Reserved.
 *
 *  Contributor(s): Markus L. Noga <noga@inrialpes.fr>
 *                  Lou Sortman <lou (at) sunsite (dot) unc (dot) edu>
 */

#include "direct-motor.h"
#include "sys/h8.h"
#include "sys/irq.h"

#ifndef NO_DIRECT_MOTOR
  
///////////////////////////////////////////////////////////////////////////////
//
// Variables
//
///////////////////////////////////////////////////////////////////////////////

//! motor drive patterns
/*! to be indexed with MotorDirections
    \sa MotorDirections
*/
const unsigned char dm_a_pattern[]={0x00,0x80,0x40,0xc0},
		    dm_b_pattern[]={0x00,0x08,0x04,0x0c},
		    dm_c_pattern[]={0x00,0x02,0x01,0x03};

MotorState dm_a,                  //!< motor A state
           dm_b,                  //!< motor B state
           dm_c;                  //!< motor C state

        
///////////////////////////////////////////////////////////////////////////////
//
// Functions
//
///////////////////////////////////////////////////////////////////////////////

//! direct motor output handler
/*! called by system timer in the 16bit timer OCIA irq
*/
extern void dm_handler(void);

__asm__("
.text
.align 1
.global _dm_handler
_dm_handler:
		; r6 saved by ROM
                ; r0 saved by systime_handler
    		
		sub.w	r6,r6			; r6l is output

                ; we simultaneously load delta (r0h) and sum (r0l)
                ; this depends on byte order, but the H8 will stay MSB
                ; and the resulting code is efficient and compact.
                
                ; motor A
                
                mov.w   @_dm_a,r0
                add.b   r0h,r0l                 ; add delta to sum
                bcc     0f                      ; sum overflow?
		  mov.b	@_dm_a+2,r6h            ; -> output drive pattern
		  or.b	r6h,r6l
              0:mov.b   r0l,@_dm_a+1            ; save sum
                                                ; (clears overflow flag)

                ; motor B
                
                mov.w   @_dm_b,r0
                add.b   r0h,r0l                 ; add delta to sum
                bcc     1f                      ; sum overflow?
		  mov.b	@_dm_b+2,r6h            ; -> output drive pattern
		  or.b	r6h,r6l
              1:mov.b   r0l,@_dm_b+1            ; save sum
                                                ; (clears overflow flag)

                ; motor C
                
                mov.w   @_dm_c,r0
                add.b   r0h,r0l                 ; add delta to sum
                bcc     2f                      ; sum overflow?
		  mov.b	@_dm_c+2,r6h            ; -> output drive pattern
		  or.b	r6h,r6l
              2:mov.b   r0l,@_dm_c+1            ; save sum
                                                ; (clears overflow flag)

		; driver chip
                  
		mov.b	r6l,@0xf000:16		; output motor waveform
		
		rts		
	");
	
		
//! initialize motors
//
void dm_init(void) {
	dm_shutdown();				// shutdown hardware
}


//! shutdown motors
//
void dm_shutdown(void) {
	RCX_MOTORS=0x00;			// shutdown hardware
        
	motor_a_dir(off);			// initialize driver data
	motor_b_dir(off);
	motor_c_dir(off);
	
	motor_a_speed(MIN_SPEED);
	motor_b_speed(MIN_SPEED);
	motor_c_speed(MIN_SPEED);
}

#endif
