/*! \file   direct-sound.c
    \brief  direct sound access
    \author Markus L. Noga <noga@inrialpes.fr>
*/
    
/*
 *  The contents of this file are subject to the Mozilla Public License
 *  Version 1.0 (the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *  http://www.mozilla.org/MPL/
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *
 *  The Original Code is legOS code, released March 19, 1999.
 *
 *  The Initial Developer of the Original Code is Markus L. Noga.
 *  Portions created by Markus L. Noga are Copyright (C) 1999
 *  Markus L. Noga. All Rights Reserved.
 *
 *  Contributor(s): Markus L. Noga <noga@inrialpes.fr>
 *                  Michael Nielsen <mic@daimi.aau.dk>
 */

#include "direct-sound.h"
#include "sys/bitops.h"
#include "sys/h8.h"
#include "sys/irq.h"

#ifndef NO_DIRECT_SOUND

///////////////////////////////////////////////////////////////////////////////
//
// Variables
//
///////////////////////////////////////////////////////////////////////////////

volatile unsigned *ds_current;			//!< current byte playing & status
volatile unsigned char *ds_buf_ptr;		//!< ptr to next byte to play
unsigned char *ds_buf_start,*ds_buf_end;	//!< ptrs to start/end of sample
unsigned ds_loop_flag;				//!< flag to loop output.
						/*< \warning unused! */

///////////////////////////////////////////////////////////////////////////////
//
// Functions
//
///////////////////////////////////////////////////////////////////////////////

//! the sound IRQ handler
//
extern void ds_switcher(void);

__asm__("
.text		
_ds_switcher:
		; r6 saved by ROM
		
		mov.w	@_ds_current,r6		; read next byte
		bld	#7,r6l			; highest bit is next state
		bst	#1,@0xc9:8		; 1 -> out 1 on Compare/Match A
		bist	#0,@0xc9:8		; 0 -> out 0 on Compare/Match A
		
	        bclr	#6,@0xc9:8		; clear compare match A flag

		shll	r6l
		dec	r6h
		beq	0f			; last bit -> next byte
		
		  mov.w r6,@_ds_current		; save ptr
		  rts				; return
		  
	      0:
		mov.w	r0,@-r7			; save r0
		
		mov.w   @_ds_buf_ptr,r0		; buffer ptr  -> r0
		mov.b   @r0+,r6l		; read next byte
		mov.b	#8,r6h			; setup counter.

		mov.w	r6,@_ds_current		; save for next round
		
		mov.w	@_ds_buf_end,r6		; end of buffer reached?
		cmp	r0,r6
		bne	1f
		
		  mov.w	@_ds_buf_start,r0	; -> start over
		  
	      1:
	        mov.w	r0,@_ds_buf_ptr		; save ptr
		
		mov.w	@r7+,r0			; restore r0
	      	
		; r6 will be restored by ROM
		
		rts				; return 
	");

	
//! start playing sound.
/*!
   \param sample an 1 bit / 8 kHz sample. util/sample-convert.pl will perform 
                 a conversion for standard wav files.
   \param length sample length in bytes
      
   \warning output sounds horrible
   \warning cycles not implemented.
*/
void ds_play(unsigned char *sample, unsigned length) {
	unsigned b;

	ds_buf_start=ds_buf_ptr=sample;
	ds_buf_end=sample+length;

#ifdef NO_SYSTEM_TIME	
	T_IER&=~TIER_ENABLE_ICA;		// turn off ROM handler
#endif
        	
	*((unsigned char*)T0_CR)  =0x00;	// timer 1 off
	*((unsigned char*)T0_CNT) =0x00;	// counter reset
	*((unsigned char*)T0_CORA)=255;		// set const A 
	
	T0_IRQA=&ds_switcher;			// set IRQ handler

	// compare B doesn't affect output
	bit_clear(T0_CSR,0x3);
	bit_clear(T0_CSR,0x2);
	
	// 0x01 + STCR0=0 : Clock / 8 -> ~2 MhZ
	// 2 MHz / 8000 Hz sampling rate = 250 -> value for count A
	bit_clear(T0_CSR,CSR_MATCH_A);
	bit_clear(STCR,0x0);		
	*((unsigned char*) T0_CR)=CR_ENABLE_IRQA | CR_CLEAR_ON_A | 0x01;// go!
}

//! stop playing sound
//
void ds_stop(void) {
	*((unsigned char*) T0_CR)=0x00;		// irq & counting off
}

#endif
