/*! \file   systime.c
    \brief  system time services
    \author Markus L. Noga <noga@inrialpes.fr>
*/
    
/*
 *  The contents of this file are subject to the Mozilla Public License
 *  Version 1.0 (the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *  http://www.mozilla.org/MPL/
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *
 *  The Original Code is legOS code, released March 19, 1999.
 *
 *  The Initial Developer of the Original Code is Markus L. Noga.
 *  Portions created by Markus L. Noga are Copyright (C) 1999
 *  Markus L. Noga. All Rights Reserved.
 *
 *  Contributor(s): Markus L. Noga <noga@inrialpes.fr>
 *                  David Van Wagner <davevw@alumni.cse.ucsc.edu>
 */

#include "config.h"

#ifndef NO_SYSTEM_TIME

#include "sys/time.h"
#include "sys/h8.h"
#include "sys/irq.h"
#include "direct-motor.h"

///////////////////////////////////////////////////////////////////////////////
//
// Global Variables
//
///////////////////////////////////////////////////////////////////////////////

//! current system time in ms
/*! \warning This is a 32 bit value which will overflow after 49.7 days
             of continuous operation.
*/
volatile time_t sys_time;

///////////////////////////////////////////////////////////////////////////////
//
// Internal Variables
//
///////////////////////////////////////////////////////////////////////////////

#ifndef NO_TASK_MANAGEMENT
unsigned char tm_timeslice;                     //!< task time slice
volatile unsigned char tm_current_slice;        //!< current time remaining

void* tm_switcher_vector;                       //!< pointer to task switcher
#endif


///////////////////////////////////////////////////////////////////////////////
//
// Functions
//
///////////////////////////////////////////////////////////////////////////////

//! system time handler for the 16bit timer OCIA irq
/*! this is the pulse of the system. 
    task switcher and motor driver calls are initiated here.
*/
extern void systime_handler(void);

__asm__("
.text
.align 1
.global _systime_handler
.global _systime_tm_return
_systime_handler:
		; r6 saved by ROM
    
                push r0                         ; both motors & task
                                                ; switcher need this reg.
  
                ; increment system timer
                
                mov.w @_sys_time+2,r6           ; LSW -> r6
                add.b #0x1,r6l                  ; 16 bit: add 1
                addx  #0x0,r6h
                mov.w r6,@_sys_time+2
                bcc 0f                          ; speedup for 65535 cases
                  
                  mov.w @_sys_time,r6           ; MSW -> r6
                  addx  #0x0,r6l
                  addx  #0x0,r6h
                  mov.w r6,@_sys_time
              0:
        "
#ifndef NO_DIRECT_MOTOR
        "
                jsr _dm_handler                 ; call motor driver
        "
#endif

#ifndef NO_TASK_MANAGEMENT
        "                
                mov.b @_tm_current_slice,r6l
                dec r6l
                bne 1f                          ; timeslice elapsed?
                
                  mov.w @_tm_switcher_vector,r6
                  jsr @r6                       ; call task switcher
_systime_tm_return:                  
                  mov.b @_tm_timeslice,r6l      ; new timeslice
                  
              1:mov.b r6l,@_tm_current_slice
        "
#endif
        "
                pop r0
                
		bclr	#3,@0x91:8		; reset compare A IRQ flag
	
		rts		
	");
	
		
//! initialize system timer
/*! task switcher initialized to empty handler
    motors turned off
*/
void systime_init(void) {
  systime_shutdown();			// shutdown hardware

  sys_time=0l;                          // init timer

#ifndef NO_TASK_MANAGEMENT
  tm_current_slice=tm_timeslice=TM_DEFAULT_SLICE;
  tm_switcher_vector=RTS_INSTRUCTION;   // empty handler
#endif
  
#ifndef NO_DIRECT_MOTOR
  dm_shutdown();        
#endif

  // configure 16-bit timer compare A IRQ 
  // to occur every 1 ms, hook and enable it.
  //
  T_CSR =TCSR_OCA | TCSR_RESET_ON_A;
  T_CR  =TCR_CLOCK_32;                  // 500 kHz @ PCLOCK=16MHz
  T_OCR&=~TOCR_OCRB;
  T_OCRA =500;                            
  
  OCIA_IRQ=&systime_handler;
  T_IER|=TIER_ENABLE_OCA;
}

//! shutdown system timer
/*! will also stop task switching and motors.
*/
void systime_shutdown(void) {
  T_IER&=~TIER_ENABLE_OCA;		// unhook compare A IRQ
}

#ifndef NO_TASK_MANAGEMENT
//! set task switcher vector
/*! \param switcher the switcher
*/
void systime_set_switcher(void* switcher) {
  tm_switcher_vector=switcher;
}

//! set multitasking timeslice in ms
/*! \param slice the timeslice. must be at least 5ms.
*/
void systime_set_timeslice(unsigned char slice) {
  if(slice>5) {                        // some minimum value
    tm_timeslice=slice;
    if(tm_current_slice>tm_timeslice)
      tm_current_slice=tm_timeslice;
  }
} 
 
#endif

#endif // NO_SYSTEM_TIME
